// stack.h - implementation of a template Stack class
//           This version uses accessor functions instead of friends

// Note: cannot be broken off to a stack.cpp implementation file
//       due to longstanding bug(s) in the Visual C++ compiler

#ifndef STACK_H
#define STACK_H

#include <iostream>
#include <string>

using namespace std;

template <class T>
class StackFrame {
  private:
    T data;
    StackFrame *link;

  public:
    StackFrame(T data, StackFrame* link = NULL) : data(data), link(link) {}
    T getData() { return data; }
    StackFrame* getLink() { return link; }
    void setLink(StackFrame* link) { this->link = link; }
};

template <class T>
class Stack {
  private:
    StackFrame<T>* top;

  public:    

    Stack() {
	top = NULL;
    }

    Stack(const Stack<T>& aStack) {
	if (aStack.top == NULL)
	    top = NULL;
	else {
	    StackFrame<T>* ptr = aStack.top;
	    StackFrame<T>* end = new StackFrame<T>(ptr->getData());
	    top = end;
	    ptr = ptr->getLink();

	    while (ptr != NULL) {
		end->setLink(new StackFrame<T>(ptr->getData()));
		end = end->getLink();
		ptr = ptr->getLink();
	    }
	}
    }

    ~Stack() {
	while (! empty())
	    pop();
    }

    void push(T data) {
	top = new StackFrame<T>(data, top);
    }

    T pop() {
	if (empty()) {
	    cerr << "Error: attempted pop of empty stack" << endl;
	    exit(1);
	}
	T data = top->getData();
	StackFrame<T>* ptr = top;
	top = top->getLink();
	delete ptr;
	return data;
    }

    bool empty() const {
	return (top == NULL);
    }

    friend ostream& operator <<(ostream& outs, Stack& aStack) {
        outs << "(";
	for (StackFrame<T>* ptr = aStack.top; ptr != NULL; ptr = ptr->getLink())
	    outs << " " << ptr->getData();
        outs << " )";
        return outs;
    }

};

#endif
